//----------------------------------------------------------------------
//
//  Software Watchdog Driver for ADNP/1520
//
//----------------------------------------------------------------------
//  AUTHOR  : Marco Hasewinkel (mha@ssv)
//  EMAIL   : <info@dilnetpc.com> <m.hasewinkel@web.de>
//  DATE    : 14.10.2005
//  VERSION : 1.00
//  LICENSE : GPL
//  COPYRIGHT (C) 2005 SSV EMBEDDED SYSTEMS GmbH
//----------------------------------------------------------------------
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//----------------------------------------------------------------------
#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>

#define SSVSWD_INTERVAL (HZ/2)
#define SSVSWD_TIMEOUT 30 /* default timeout 30s */
#define SSVSWD_MODNAME "ssvswd"

static void ssvswd_timerfunction(unsigned long);
static unsigned long ssvswd_timercount;
static struct timer_list ssvswd_timer;
volatile unsigned short *ssvswd_flash;
static unsigned long ssvswd_is_open;
static int ssvswd_expect_close, timeout = 0;

#ifdef CONFIG_WATCHDOG_NOWAYOUT
static int nowayout = 1;
#else
static int nowayout = 0;
#endif

MODULE_PARM(nowayout,"i");
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started (default=CONFIG_WATCHDOG_NOWAYOUT)");

MODULE_PARM(timeout,"i");
MODULE_PARM_DESC(timeout, "Watchdog timeout in seconds");

static void ssvswd_resetsystem(void)
{
    cli();

    *ssvswd_flash = 0xff; // read reset
    *ssvswd_flash = 0x50; // clear status register
    *ssvswd_flash = 0xff; // read reset

    outb(0x3, 0x92); // system reset

    for (;;); /* endless */
}

static void ssvswd_timerfunction(unsigned long data)
{
    if (ssvswd_timercount) {
	ssvswd_timercount--;
    }
    else {
	ssvswd_resetsystem();
    }

    ssvswd_timer.expires = jiffies + SSVSWD_INTERVAL;
    add_timer(&ssvswd_timer);
}

static void ssvswd_startup(void)
{
    ssvswd_timercount = timeout;

    if ((nowayout) && (MOD_IN_USE > 1)) {
	del_timer(&ssvswd_timer);
    }
    ssvswd_timer.expires = jiffies + SSVSWD_INTERVAL;	
    add_timer(&ssvswd_timer);

    printk(SSVSWD_MODNAME ": Watchdog timer is now enabled.\n");
}

static void ssvswd_turnoff(void)
{
    if (!nowayout) {
	del_timer(&ssvswd_timer);
	printk(SSVSWD_MODNAME ": Watchdog timer is now disabled...\n");
    }
}

static ssize_t ssvswd_fop_write(struct file * file, const char * buf, size_t count, loff_t * ppos)
{
    /* no seek */
    if(ppos != &file->f_pos)
	return -ESPIPE;

    if(count) {
	size_t ofs;

	ssvswd_expect_close = 0;

	for(ofs = 0; ofs != count; ofs++) {
	    char c;
	    if (get_user(c, buf + ofs))
		return -EFAULT;
	    if(c == 'V')
		ssvswd_expect_close = 1;
	}
	    ssvswd_timercount = timeout;
	    return 1;
	}

    return 0;
}

static int ssvswd_fop_open(struct inode * inode, struct file * file)
{
    switch(MINOR(inode->i_rdev)) {
	case WATCHDOG_MINOR:
	    if(test_and_set_bit(0, &ssvswd_is_open))
		return -EBUSY;

	    ssvswd_startup();
	    if (nowayout) {
		MOD_INC_USE_COUNT;
	    }
	    return 0;

	default:
	    return -ENODEV;
    }
}

static int ssvswd_fop_close(struct inode * inode, struct file * file)
{
    if(MINOR(inode->i_rdev) == WATCHDOG_MINOR) {
	if(ssvswd_expect_close) {
	    ssvswd_turnoff();
	}
	else {
			del_timer(&ssvswd_timer);
			printk(SSVSWD_MODNAME ": device file closed unexpectedly. Reset System !\n");
			ssvswd_resetsystem();
	}
    }

    clear_bit(0, &ssvswd_is_open);

    return 0;
}

static long long ssvswd_fop_llseek(struct file *file, long long offset, int origin)
{
    return -ESPIPE;
}

static int ssvswd_fop_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
    static struct watchdog_info ident = {
	WDIOF_MAGICCLOSE,
	1,
	"softwatchdog"
    };
	
    switch(cmd) {
	default:
	    return -ENOIOCTLCMD;

	case WDIOC_GETSUPPORT:
	    return copy_to_user((struct watchdog_info *)arg, &ident, sizeof(ident))?-EFAULT:0;

	case WDIOC_KEEPALIVE:
	    ssvswd_timercount = timeout;
	    return 0;
    }
}

static struct file_operations ssvswd_fops = {
    owner:	THIS_MODULE,
    llseek:	ssvswd_fop_llseek,
    write:	ssvswd_fop_write,
    open:	ssvswd_fop_open,
    release:	ssvswd_fop_close,
    ioctl:	ssvswd_fop_ioctl
};

static struct miscdevice ssvswd_miscdev = {
    WATCHDOG_MINOR,
    "watchdog",
    &ssvswd_fops
};

static int ssvswd_notify_sys(struct notifier_block *this, unsigned long code, void *unused)
{
    if(code==SYS_DOWN || code==SYS_HALT) 
	ssvswd_turnoff();

    return NOTIFY_DONE;
}
 
static struct notifier_block ssvswd_notifier = {
    ssvswd_notify_sys,
    0,
    0
};

static void __exit ssvswd_unload(void)
{
    ssvswd_turnoff();

    misc_deregister(&ssvswd_miscdev);
    iounmap((void *)ssvswd_flash);
    unregister_reboot_notifier(&ssvswd_notifier);
}

static int __init ssvswd_init(void)
{
    int rc = -EBUSY;

    /* range for timeout 1s - 3600s */
    if ((timeout > 0) && (timeout < 3601)) {
	timeout = timeout * HZ / SSVSWD_INTERVAL;
    }
    else {
	timeout = (SSVSWD_TIMEOUT * HZ / SSVSWD_INTERVAL);
    }

    init_timer(&ssvswd_timer);
    ssvswd_timer.function = ssvswd_timerfunction;
    ssvswd_timer.data = 0;

    rc = misc_register(&ssvswd_miscdev);
    if (rc) goto err_out_region2;

    rc = register_reboot_notifier(&ssvswd_notifier);
    if (rc) goto err_out_miscdev;

    ssvswd_flash = ioremap_nocache(0x10000000, 0x1000);

    printk(KERN_INFO SSVSWD_MODNAME ": Software Watchdog initialised. Timeout: %d sec\n", timeout * SSVSWD_INTERVAL / HZ);

    return 0;

err_out_miscdev:
    misc_deregister(&ssvswd_miscdev);
err_out_region2:
    return rc;
}

module_init(ssvswd_init);
module_exit(ssvswd_unload);

MODULE_AUTHOR("MHA@SSV");
MODULE_DESCRIPTION("Driver for ADNP/1520 Software Watchdog");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;
